"use strict";
;
;
;
;
;
;
;
/// <reference path="common.ts" />
/// <reference path="system.ts" />
var MaximizeFlags;
(function (MaximizeFlags) {
    MaximizeFlags[MaximizeFlags["Neither"] = 0] = "Neither";
})(MaximizeFlags || (MaximizeFlags = {}));
;
var GnomeSystem;
(function (GnomeSystem) {
    var _a = imports.gi, GLib = _a.GLib, Meta = _a.Meta;
    var Main = imports.ui.main;
    GnomeSystem.Clutter = imports.gi.Clutter;
    GnomeSystem.Cairo = imports.cairo;
    function newClutterActor() { return new GnomeSystem.Clutter.Actor(); }
    GnomeSystem.newClutterActor = newClutterActor;
    function newClutterCanvas() { return new GnomeSystem.Clutter.Canvas(); }
    GnomeSystem.newClutterCanvas = newClutterCanvas;
    function newClutterColor(components) {
        return new GnomeSystem.Clutter.Color(components);
    }
    GnomeSystem.newClutterColor = newClutterColor;
    function numWorkspaces() {
        return global.workspace_manager.get_n_workspaces();
    }
    GnomeSystem.numWorkspaces = numWorkspaces;
    function workspaceIndex() {
        return global.workspace_manager.get_active_workspace_index();
    }
    GnomeSystem.workspaceIndex = workspaceIndex;
    function activateWorkspace(idx) {
        var ws = global.workspace_manager.get_workspace_by_index(idx);
        ws.activate(global.get_current_time());
    }
    GnomeSystem.activateWorkspace = activateWorkspace;
    function moveWindowToWorkspace(win, idx) {
        var ws = global.workspace_manager.get_workspace_by_index(idx);
        win.change_workspace_by_index(idx, false);
        ws.activate_with_focus(win, global.get_current_time());
    }
    GnomeSystem.moveWindowToWorkspace = moveWindowToWorkspace;
    function stableSequence(win) {
        return win.get_stable_sequence();
    }
    GnomeSystem.stableSequence = stableSequence;
    function windowTitle(win) {
        return win.get_title();
    }
    GnomeSystem.windowTitle = windowTitle;
    function rect(metaRect) {
        return {
            pos: { x: metaRect.x, y: metaRect.y },
            size: { x: metaRect.width, y: metaRect.height }
        };
    }
    function size(metaRect) {
        return { x: metaRect.width, y: metaRect.height };
    }
    function windowRect(win) {
        var globalRect = rect(win.get_frame_rect());
        return Rect.move(globalRect, Point.scaleConstant(-1, workspaceOffset(win)));
    }
    GnomeSystem.windowRect = windowRect;
    function minimize(win) {
        win.minimize();
    }
    GnomeSystem.minimize = minimize;
    function unminimize(win) {
        win.unminimize();
    }
    GnomeSystem.unminimize = unminimize;
    function moveResize(win, workspaceRect) {
        if (win.get_maximized() !== MaximizeFlags.Neither) {
            unmaximize(win);
        }
        var rect = Rect.move(workspaceRect, workspaceOffset(win));
        win.move_resize_frame(true, rect.pos.x, rect.pos.y, rect.size.x, rect.size.y);
    }
    GnomeSystem.moveResize = moveResize;
    function currentWindow() {
        return global.display['focus-window'];
    }
    GnomeSystem.currentWindow = currentWindow;
    function workspaceOffset(win) {
        return rect(win.get_work_area_current_monitor()).pos;
    }
    GnomeSystem.workspaceOffset = workspaceOffset;
    function translateEventCoordinates(point, win) {
        return Point.subtract(point, workspaceOffset(win));
    }
    GnomeSystem.translateEventCoordinates = translateEventCoordinates;
    function workspaceArea(win) {
        return size(win.get_work_area_current_monitor());
    }
    GnomeSystem.workspaceArea = workspaceArea;
    function unmaximize(win) {
        win.unmaximize(Meta.MaximizeFlags.VERTICAL | Meta.MaximizeFlags.HORIZONTAL);
    }
    GnomeSystem.unmaximize = unmaximize;
    function maximize(win) {
        win.maximize(Meta.MaximizeFlags.VERTICAL | Meta.MaximizeFlags.HORIZONTAL);
    }
    GnomeSystem.maximize = maximize;
    function getMaximized(win) {
        return win.get_maximized() !== MaximizeFlags.Neither;
    }
    GnomeSystem.getMaximized = getMaximized;
    var VISIBLE_WINDOW_TYPES = [
        Meta.WindowType.NORMAL,
        Meta.WindowType.DIALOG,
        Meta.WindowType.UTILITY,
    ];
    function listWindows(win) {
        var display = (win == null) ? global.display : win.get_display();
        var screenNo = (win == null) ? display.get_current_monitor() : win.get_monitor();
        var windows = display.get_workspace_manager().get_active_workspace().list_windows();
        return windows.filter(function (w) {
            return (w.get_monitor() == screenNo
                && VISIBLE_WINDOW_TYPES.indexOf(w.get_window_type()) !== -1);
        });
    }
    function visibleWindows() {
        var current = currentWindow();
        var windows = listWindows(current).filter(function (w) {
            return !w.minimized;
        });
        return [current, windows];
    }
    GnomeSystem.visibleWindows = visibleWindows;
    function minimizedWindows() {
        return listWindows().filter(function (w) {
            return w.minimized;
        });
    }
    GnomeSystem.minimizedWindows = minimizedWindows;
    function activate(win) {
        Main.activateWindow(win, global.get_current_time());
    }
    GnomeSystem.activate = activate;
    function activateLater(win) {
        GLib.idle_add(function () {
            Main.activateWindow(win, global.get_current_time());
            return false;
        });
    }
    GnomeSystem.activateLater = activateLater;
    function setWindowHidden(win, hidden) {
        win.get_compositor_private().set_opacity(hidden ? 0 : 255);
    }
    GnomeSystem.setWindowHidden = setWindowHidden;
})(GnomeSystem || (GnomeSystem = {}));
/// <reference path="common.ts" />
/// <reference path="gnome_shell.ts" />
var Settings;
(function (Settings) {
    var Gio = imports.gi.Gio;
    var ExtensionUtils = imports.misc.extensionUtils;
    var Ext = ExtensionUtils.getCurrentExtension();
    var SCHEMA_ROOT = 'org.gnome.shell.extensions.net.gfxmonk.slinger';
    var KEYBINDINGS = SCHEMA_ROOT + '.keybindings';
    function get_local_gsettings(schema_path) {
        var GioSSS = Gio.SettingsSchemaSource;
        var schemaDir = Ext.dir.get_child('schemas');
        var schemaSource = GioSSS.new_from_directory(schemaDir.get_path(), GioSSS.get_default(), false);
        var schemaObj = schemaSource.lookup(schema_path, true);
        if (!schemaObj) {
            throw new Error('Schema ' + schema_path +
                ' could not be found for extension ' +
                Ext.metadata.uuid);
        }
        return new Gio.Settings({ settings_schema: schemaObj });
    }
    ;
    var Keybindings = /** @class */ (function () {
        function Keybindings() {
            this.settings = get_local_gsettings(KEYBINDINGS);
        }
        return Keybindings;
    }());
    Settings.Keybindings = Keybindings;
    ;
})(Settings || (Settings = {}));
function assert(x, msg) {
    if (!x) {
        throw new Error(msg ? ("Assertion failed: " + msg) : "Assertion failed");
    }
    return x;
}
/// <reference path="common.ts" />
function p(msg) {
    log('[slinger]:' + msg);
}
function dump(obj) {
    p(JSON.stringify(obj));
}
var MenuSelection = /** @class */ (function () {
    function MenuSelection(ring, index, splitMode) {
        this.ring = ring;
        this.index = index;
        this.splitMode = splitMode;
    }
    MenuSelection.prototype.eq = function (b) {
        return this.ring == b.ring && this.index == b.index && this.splitMode == b.splitMode;
    };
    // like eqTo, but ignores splitMode
    MenuSelection.prototype.eqPosition = function (ring, location) {
        return this.eqTo(ring, location, this.splitMode);
    };
    MenuSelection.prototype.eqTo = function (ring, location, splitMode) {
        return this.ring == ring && this.index == location && this.splitMode == splitMode;
    };
    MenuSelection.prototype.Inner = function (sel) {
        return MenuSelection.Inner(sel, this.splitMode);
    };
    MenuSelection.prototype.Outer = function (sel) {
        return MenuSelection.Outer(sel, this.splitMode);
    };
    return MenuSelection;
}());
(function (MenuSelection) {
    function None(splitMode) {
        return new MenuSelection(0 /* NONE */, 0, splitMode);
    }
    MenuSelection.None = None;
    function Inner(sel, splitMode) {
        return new MenuSelection(1 /* INNER */, sel, splitMode);
    }
    MenuSelection.Inner = Inner;
    function Outer(sel, splitMode) {
        return new MenuSelection(2 /* OUTER */, sel, splitMode);
    }
    MenuSelection.Outer = Outer;
})(MenuSelection || (MenuSelection = {}));
var MathUtil;
(function (MathUtil) {
    function between(min, x, max) {
        // p('between('+min+', '+x+', '+max+')');
        if (min > x)
            return min;
        if (max < x)
            return max;
        return x;
    }
    MathUtil.between = between;
})(MathUtil || (MathUtil = {}));
/// <reference path="common.ts" />
/// <reference path="logging.ts" />
/// <reference path="menu.ts" />
/// <reference path="menu_selection.ts" />
/// <reference path="math.ts" />
var Preview;
(function (Preview) {
    var MANIPULATION_SCALE = 2.2;
    var MINIMUM_SIZE = 20;
    var ONE_THIRD = 1.0 / 3.0;
    var TWO_THIRD = 2.0 / 3.0;
    function oppose(loc) {
        return (loc + 4) % 8; // Magic!
    }
    Preview.oppose = oppose;
    var LayoutPreview = /** @class */ (function () {
        function LayoutPreview(Sys, size, windowRect, win) {
            this.Sys = Sys;
            this.size = size;
            this.ui = Sys.newClutterActor();
            this.ui.set_background_color(Sys.newClutterColor({
                red: 80,
                green: 158,
                blue: 255,
                alpha: 125
            }));
            this.resizeCorner = null;
            this.selection = MenuSelection.None(null);
            this.windowRect = windowRect;
            this.win = win;
            this.ui.hide();
            this.windowHidden = null;
        }
        LayoutPreview.prototype.selectOuter = function (loc, splitMode) {
            var size = this.size;
            switch (splitMode) {
                case 0 /* FOUR */:
                    switch (loc) {
                        case 0 /* LEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: 0.5, y: 1 }, size)
                            };
                        case 1 /* TOPLEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scaleConstant(0.5, size)
                            };
                        case 2 /* TOP */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: 1, y: 0.5 }, size)
                            };
                        case 3 /* TOPRIGHT */:
                            return {
                                pos: Point.scale({ x: 0.5, y: 0 }, size),
                                size: Point.scaleConstant(0.5, size)
                            };
                        case 4 /* RIGHT */:
                            return {
                                pos: Point.scale({ x: 0.5, y: 0 }, size),
                                size: Point.scale({ x: 0.5, y: 1 }, size)
                            };
                        case 5 /* BOTTOMRIGHT */:
                            return {
                                pos: Point.scaleConstant(0.5, size),
                                size: Point.scaleConstant(0.5, size)
                            };
                        case 6 /* BOTTOM */:
                            return {
                                pos: Point.scale({ x: 0, y: 0.5 }, size),
                                size: Point.scale({ x: 1, y: 0.5 }, size)
                            };
                        case 7 /* BOTTOMLEFT */:
                            return {
                                pos: Point.scale({ x: 0, y: 0.5 }, size),
                                size: Point.scaleConstant(0.5, size)
                            };
                    }
                case 1 /* SIX */:
                    switch (loc) {
                        case 0 /* LEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: ONE_THIRD, y: 1 }, size)
                            };
                        case 1 /* TOPLEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size)
                            };
                        case 2 /* TOP */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size)
                            };
                        case 3 /* TOPRIGHT */:
                            return {
                                pos: Point.scale({ x: TWO_THIRD, y: 0 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size)
                            };
                        case 4 /* RIGHT */:
                            return {
                                pos: Point.scale({ x: TWO_THIRD, y: 0 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 1 }, size)
                            };
                        case 5 /* BOTTOMRIGHT */:
                            return {
                                pos: Point.scale({ x: TWO_THIRD, y: 0.5 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size)
                            };
                        case 6 /* BOTTOM */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size)
                            };
                        case 7 /* BOTTOMLEFT */:
                            return {
                                pos: Point.scale({ x: 0, y: 0.5 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size)
                            };
                    }
            }
            return null;
        };
        LayoutPreview.prototype.selectInner = function (sel, splitMode) {
            switch (splitMode) {
                case 0 /* FOUR */:
                    switch (sel) {
                        case 2 /* TOP */:
                            return {
                                pos: Point.ZERO,
                                size: this.size
                            };
                        case 6 /* BOTTOM */:
                            return null;
                    }
                    break;
                case 1 /* SIX */:
                    switch (sel) {
                        case 1 /* TOPLEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: TWO_THIRD, y: 1 }, this.size)
                            };
                        case 3 /* TOPRIGHT */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0 }, this.size),
                                size: Point.scale({ x: TWO_THIRD, y: 1 }, this.size)
                            };
                        case 6 /* BOTTOM */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0 }, this.size),
                                size: Point.scale({ x: ONE_THIRD, y: 1 }, this.size)
                            };
                    }
                    break;
            }
            return null;
        };
        LayoutPreview.prototype.trackMouse = function (prevMode, origin) {
            if (prevMode == 0 /* MENU */) {
                switch (this.selection.ring) {
                    case 1 /* INNER */:
                        return false;
                    case 2 /* OUTER */:
                        this.resizeCorner = oppose(this.selection.index);
                        this.trackingOrigin = origin;
                        break;
                    case 0 /* NONE */:
                        this.resetTracking(origin);
                        break;
                }
            }
            else {
                this.resetTracking(origin);
            }
            this.base = this.preview; // capture whatever we have as a base
            this.updateUi();
            return true;
        };
        LayoutPreview.prototype.resetTracking = function (origin) {
            if (this.preview == null) {
                this.preview = this.windowRect;
            }
            this.base = this.preview;
            this.resizeCorner = Rect.closestCorner(this.preview, origin);
            this.trackingOrigin = origin;
            this.updateUi();
        };
        LayoutPreview.prototype.onMouseMove = function (mode, point) {
            switch (mode) {
                case 1 /* RESIZE */:
                    if (this.resizeCorner === null) {
                        return;
                    }
                    var diff = Point.subtract(point, this.trackingOrigin);
                    this.preview = LayoutPreview.applyResize(this.resizeCorner, diff, this.base, this.size);
                    // p('move diff ' + JSON.stringify(diff)
                    // 	+ ' (from origin ' + JSON.stringify(this.trackingOrigin) + ')'
                    // 	+ ' turned base ' + JSON.stringify(this.base)
                    // 	+ ' into rect ' + JSON.stringify(this.preview)
                    // );
                    break;
                case 2 /* MOVE */:
                    var diff = Point.subtract(point, this.trackingOrigin);
                    this.preview = LayoutPreview.applyMove(diff, this.base, this.size);
                    // p('move diff ' + JSON.stringify(diff)
                    // 	+ ' (from origin ' + JSON.stringify(this.trackingOrigin) + ')'
                    // 	+ ' turned base ' + JSON.stringify(this.base)
                    // 	+ ' into rect ' + JSON.stringify(this.preview)
                    // );
                    break;
                default:
                    return;
            }
            this.updateUi();
        };
        LayoutPreview.prototype.applyManipulator = function (fn) {
            if (this.preview === null)
                return;
            this.preview = fn(this.preview, this.size);
            this.updateUi();
        };
        LayoutPreview.applyMove = function (diff, base, bounds) {
            var ret = Rect.copy(base);
            var scaled = Point.scaleConstant(MANIPULATION_SCALE, diff);
            ret.pos.x = Math.max(0, Math.min(ret.pos.x + scaled.x, bounds.x - ret.size.x));
            ret.pos.y = Math.max(0, Math.min(ret.pos.y + scaled.y, bounds.y - ret.size.y));
            return ret;
        };
        LayoutPreview.applyResize = function (location, diff, base, bounds) {
            var ret = Rect.copy(base);
            var scaled = Point.scaleConstant(MANIPULATION_SCALE, diff);
            function moveNear(axis) {
                // minimum diff is enough to bring this edge to 0 (i.e. invert the current pos)
                // maximum diff is enough to bring this edge to ther other side of this rect, minus MINIMUM_SIZE
                // p('moveNear['+axis+']');
                var diff = MathUtil.between(-ret.pos[axis], scaled[axis], ret.size[axis] - MINIMUM_SIZE);
                ret.pos[axis] += diff;
                ret.size[axis] -= diff;
            }
            function moveFar(axis) {
                // minimum diff is enough to bring this edge to the other side of this rect, plus MINIMUM_SIZE
                // maximum diff is enough to bring this edge to the right bounds
                // p('moveFar['+axis+']');
                var diff = MathUtil.between(MINIMUM_SIZE - ret.size[axis], scaled[axis], bounds[axis] - ret.pos[axis] - ret.size[axis]);
                ret.size[axis] += diff;
            }
            switch (location) {
                case 0 /* LEFT */:
                    moveNear("x" /* x */);
                    break;
                case 1 /* TOPLEFT */:
                    moveNear("x" /* x */);
                    moveNear("y" /* y */);
                    break;
                case 2 /* TOP */:
                    moveNear("y" /* y */);
                    break;
                case 3 /* TOPRIGHT */:
                    moveNear("y" /* y */);
                    moveFar("x" /* x */);
                    break;
                case 4 /* RIGHT */:
                    moveFar("x" /* x */);
                    break;
                case 5 /* BOTTOMRIGHT */:
                    moveFar("x" /* x */);
                    moveFar("y" /* y */);
                    break;
                case 6 /* BOTTOM */:
                    moveFar("y" /* y */);
                    break;
                case 7 /* BOTTOMLEFT */:
                    moveNear("x" /* x */);
                    moveFar("y" /* y */);
                    break;
                default:
                    throw new Error("unknown location: " + location);
            }
            return ret;
        };
        LayoutPreview.prototype.updateSelection = function (sel) {
            this.selection = sel;
            switch (sel.ring) {
                case 2 /* OUTER */:
                    this.base = this.selectOuter(sel.index, sel.splitMode);
                    break;
                case 1 /* INNER */:
                    this.base = this.selectInner(sel.index, sel.splitMode);
                    break;
                case 0 /* NONE */:
                default:
                    this.base = null;
                    break;
            }
            this.resetPreview();
        };
        LayoutPreview.prototype.setWindowHidden = function (hidden) {
            if (this.win !== null && this.windowHidden !== hidden) {
                this.windowHidden = hidden;
                this.Sys.setWindowHidden(this.win, hidden);
            }
        };
        LayoutPreview.prototype.updateUi = function () {
            if (this.preview == null) {
                this.ui.hide();
                this.setWindowHidden(false);
            }
            else {
                this.ui.set_size(this.preview.size.x, this.preview.size.y);
                this.ui.set_position(this.preview.pos.x, this.preview.pos.y);
                this.setWindowHidden(true);
                this.ui.show();
            }
        };
        LayoutPreview.prototype.resetPreview = function () {
            if (this.base == null) {
                this.preview = null;
            }
            else {
                this.preview = Rect.copy(this.base);
            }
            this.updateUi();
        };
        LayoutPreview.prototype.getRect = function () { return this.preview; };
        return LayoutPreview;
    }());
    Preview.LayoutPreview = LayoutPreview;
})(Preview || (Preview = {}));
var Point;
(function (Point) {
    var floor = Math.floor;
    function scale(scale, p) {
        return {
            x: floor(p.x * scale.x),
            y: floor(p.y * scale.y)
        };
    }
    Point.scale = scale;
    function scaleConstant(scale, p) {
        return {
            x: floor(p.x * scale),
            y: floor(p.y * scale)
        };
    }
    Point.scaleConstant = scaleConstant;
    function scaleAxis(axis, scale, p) {
        var ret = copy(p);
        ret[axis] = floor(p[axis] * scale);
        return ret;
    }
    Point.scaleAxis = scaleAxis;
    function copy(p) {
        return { x: p.x, y: p.y };
    }
    Point.copy = copy;
    function add(a, b) {
        return {
            x: a.x + b.x,
            y: a.y + b.y
        };
    }
    Point.add = add;
    function subtract(a, b) {
        return {
            x: a.x - b.x,
            y: a.y - b.y
        };
    }
    Point.subtract = subtract;
    Point.ZERO = { x: 0, y: 0 };
    function isPositive(p) {
        return p.x > 0 && p.y > 0;
    }
    Point.isPositive = isPositive;
    function eqTo(p, x, y) {
        return p.x === x && p.y === y;
    }
    Point.eqTo = eqTo;
    function ofEvent(event) {
        var _a = event.get_coords(), absx = _a[0], absy = _a[1];
        return { x: absx, y: absy };
    }
    Point.ofEvent = ofEvent;
    function magnitude(p) {
        return Math.sqrt(Math.pow(p.x, 2) + Math.pow(p.y, 2));
    }
    Point.magnitude = magnitude;
})(Point || (Point = {}));
/// <reference path="common.ts" />
/// <reference path="point.ts" />
/// <reference path="math.ts" />
var Rect;
(function (Rect) {
    function copy(r) {
        return { pos: Point.copy(r.pos), size: Point.copy(r.size) };
    }
    Rect.copy = copy;
    function area(r) {
        return r.size.x * r.size.y;
    }
    Rect.area = area;
    function isPositive(r) {
        return Point.isPositive(r.size);
    }
    Rect.isPositive = isPositive;
    function closestCorner(r, p) {
        var pow = Math.pow;
        var sqrt = Math.sqrt;
        var near = r.pos;
        var far = { x: r.pos.x + r.size.x, y: r.pos.y + r.size.y };
        var tl = sqrt(pow(p.x - near.x, 2) + pow(p.y - near.y, 2));
        var tr = sqrt(pow(p.x - far.x, 2) + pow(p.y - near.y, 2));
        var br = sqrt(pow(p.x - far.x, 2) + pow(p.y - far.y, 2));
        var bl = sqrt(pow(p.x - near.x, 2) + pow(p.y - far.y, 2));
        var corners = [tl, tr, bl, br];
        // log("Closest corners: " + JSON.stringify(corners));
        corners.sort(function (a, b) { return a - b; });
        // log("Closest corners (sorted): " + JSON.stringify(corners));
        var min = corners[0];
        if (min == tl)
            return 1 /* TOPLEFT */;
        if (min == tr)
            return 3 /* TOPRIGHT */;
        if (min == br)
            return 5 /* BOTTOMRIGHT */;
        return 7 /* BOTTOMLEFT */;
    }
    Rect.closestCorner = closestCorner;
    function move(r, p) {
        return {
            pos: Point.add(r.pos, p),
            size: r.size
        };
    }
    Rect.move = move;
    function moveWithin(rect, bounds) {
        var r = copy(rect);
        if (r.size.x > bounds.size.x)
            r.size.x = bounds.size.x;
        if (r.size.y > bounds.size.y)
            r.size.y = bounds.size.y;
        r.pos.x = MathUtil.between(bounds.pos.x, r.pos.x, (bounds.pos.x + bounds.size.x - r.size.x));
        r.pos.y = MathUtil.between(bounds.pos.y, r.pos.y, (bounds.pos.y + bounds.size.y - r.size.y));
        return r;
    }
    Rect.moveWithin = moveWithin;
    function scaleWithin(rect, size) {
        return moveWithin(rect, { pos: Point.ZERO, size: size });
    }
    Rect.scaleWithin = scaleWithin;
    function midpoint(rect) {
        return Point.add(rect.pos, Point.scaleConstant(0.5, rect.size));
    }
    Rect.midpoint = midpoint;
    function moveWithinAffinity(opts) {
        var abs = Math.abs;
        var original = opts.original, bounds = opts.bounds;
        var modified = scaleWithin(opts.modified, bounds);
        function close(a, b) {
            return abs(a - b) < opts.margin;
        }
        function axisDiff(axis) {
            var near = close(0, original.pos[axis]);
            var far = close((bounds[axis]), (original.pos[axis] + original.size[axis]));
            // p("near affinity["+axis+"] = "+near);
            // p("far affinity["+axis+"] = "+far);
            if (near && far)
                return 0; // can't do anything for a window attached to both edges
            if (near) {
                // p("maintaining near affinity");
                return -modified.pos[axis];
            }
            else if (far) {
                // p("maintaining far affinity");
                return (bounds[axis]) - (modified.pos[axis] + modified.size[axis]);
            }
            return 0;
        }
        return move(modified, {
            x: axisDiff("x" /* x */),
            y: axisDiff("y" /* y */)
        });
    }
    Rect.moveWithinAffinity = moveWithinAffinity;
    function intersection(a, b) {
        var pos = {
            x: Math.max(a.pos.x, b.pos.x),
            y: Math.max(a.pos.y, b.pos.y)
        };
        var bottomRightA = Point.add(a.pos, a.size);
        var bottomRightB = Point.add(b.pos, b.size);
        var bottomRight = {
            x: Math.min(bottomRightA.x, bottomRightB.x),
            y: Math.min(bottomRightA.y, bottomRightB.y)
        };
        var size = Point.subtract(bottomRight, pos);
        if (Point.isPositive(size)) {
            return { pos: pos, size: size };
        }
        else {
            return null;
        }
    }
    Rect.intersection = intersection;
})(Rect || (Rect = {}));
var Manipulations;
(function (Manipulations) {
    var RESIZE_SENSITIVITY = 0.1;
    var MOVE_SENSITIVITY = 0.1;
    var EDGE_AFFINITY = 30; // px
    function axialPoint(axis, magnitude) {
        switch (axis) {
            case null: return { x: magnitude, y: magnitude };
            case "x" /* x */: return { x: magnitude, y: 0 };
            case "y" /* y */: return { x: 0, y: magnitude };
        }
    }
    function scaleForWorkspace(workArea) {
        return (workArea.x + workArea.y) / 2;
    }
    function wrap(edgeAffinity, fn) {
        return function (rect, workArea) {
            var scale = scaleForWorkspace(workArea);
            var newRect = fn(rect, scale);
            if (edgeAffinity) {
                return Rect.moveWithinAffinity({
                    margin: EDGE_AFFINITY,
                    original: rect,
                    modified: newRect,
                    bounds: workArea
                });
            }
            else {
                return Rect.scaleWithin(newRect, workArea);
            }
        };
    }
    function resize(direction, axis) {
        return wrap(true, function (rect, scale) {
            var amount = RESIZE_SENSITIVITY * scale * direction;
            var diff = axialPoint(axis, amount);
            // p("scaling diff: " + JSON.stringify(diff) + " on window rect: " + JSON.stringify(rect));
            return {
                // remove half of diff from pos to keep centered
                pos: Point.add(rect.pos, Point.scaleConstant(-0.5, diff)),
                size: Point.add(rect.size, diff)
            };
        });
    }
    Manipulations.resize = resize;
    function move(direction, axis) {
        return wrap(false, function (rect, scale) {
            var amount = MOVE_SENSITIVITY * scale * direction;
            var diff = axialPoint(axis, amount);
            // p("adding diff: " + JSON.stringify(amount) + " to window rect: " + JSON.stringify(rect));
            return {
                pos: Point.add(rect.pos, diff),
                size: rect.size
            };
        });
    }
    Manipulations.move = move;
    var RectAndArea = /** @class */ (function () {
        function RectAndArea(rect) {
            this.rect = rect;
            this.area = Rect.area(rect);
        }
        RectAndArea.from = function (r) {
            return new RectAndArea(r);
        };
        RectAndArea.prototype.intersection = function (other) {
            var intersection = Rect.intersection(this.rect, other.rect);
            if (intersection == null)
                return null;
            return new RectAndArea(intersection);
        };
        RectAndArea.descendingSortOrder = function (a, b) {
            return a.area - b.area;
        };
        return RectAndArea;
    }());
    function freeRectsAround(workArea, win) {
        var bottomRight = Point.add(win.pos, win.size);
        var top = { pos: Point.ZERO, size: { x: workArea.x, y: win.pos.y } };
        var left = { pos: Point.ZERO, size: { x: win.pos.x, y: workArea.y } };
        var bottom = { pos: { x: 0, y: bottomRight.y }, size: { x: workArea.x, y: workArea.y - bottomRight.y } };
        var right = { pos: { x: bottomRight.x, y: 0 }, size: { x: workArea.x - bottomRight.x, y: workArea.y } };
        return ([top, left, bottom, right]
            .filter(Rect.isPositive)
            .map(RectAndArea.from)
            .sort(RectAndArea.descendingSortOrder));
    }
    function largestFreeRect(workArea, windows, minArea) {
        var freeWindowAreas = windows.map(function (win) { return freeRectsAround(workArea, win); });
        // log("freeWindowAreas = " + JSON.stringify(freeWindowAreas));
        var largest = null;
        function worthPursuing(candidate) {
            if (candidate == null || candidate.area < minArea)
                return false;
            return largest == null || candidate.area > largest.area;
        }
        // pick one free rect per window
        function recurse(rect, remaining) {
            // log("recursing over rect " + JSON.stringify(rect) + ", with " + remaining.length + " windows remaining");
            if (remaining.length == 0) {
                // we've picked one free rect from each window, and the intersection is nonzero
                if (worthPursuing(rect)) {
                    largest = rect;
                }
                return;
            }
            var nextSet = remaining[0];
            // TODO: we'd save allocations using indexes
            var nextRemaining = remaining.slice(1);
            nextSet.forEach(function (candidate) {
                var intersection = rect.intersection(candidate);
                // log("intersection of " + JSON.stringify(rect) + " and " + JSON.stringify(candidate) + " = " + JSON.stringify(intersection));
                if (worthPursuing(intersection)) {
                    // valid intersection which isn't already smaller than our largest candidate, keep going
                    recurse(intersection, nextRemaining);
                }
            });
        }
        recurse(new RectAndArea({ pos: Point.ZERO, size: workArea }), freeWindowAreas);
        if (largest == null || largest.area < minArea)
            return null;
        return largest.rect;
    }
    Manipulations.largestFreeRect = largestFreeRect;
})(Manipulations || (Manipulations = {}));
/// <reference path="common.ts" />
/// <reference path="logging.ts" />
/// <reference path="preview.ts" />
/// <reference path="point.ts" />
/// <reference path="rect.ts" />
/// <reference path="menu.ts" />
/// <reference path="menu_selection.ts" />
var MenuHandlers;
(function (MenuHandlers) {
    var PI = Math.PI;
    var TAO = 2 * PI;
    var floor = Math.floor;
    var ANGLE_HALF = PI;
    var ANGLE_QUARTER = ANGLE_HALF / 2;
    var ANGLE_EIGHTH = ANGLE_QUARTER / 2;
    var ANGLE_SIXTEENTH = ANGLE_EIGHTH / 2;
    var ANGLE_TWELFTH = TAO / 12.0;
    function circularIndex(sections, offset, mapFn) {
        var span = TAO / sections;
        return function (angle) {
            return mapFn(floor((angle + PI + offset) / span) % sections);
        };
    }
    function selectionOfIndex(selections) {
        return function (index) {
            return selections[index];
        };
    }
    var identity = function (x) { return x; };
    var outerSelection = circularIndex(8, ANGLE_SIXTEENTH, identity);
    var innerSelectionFour = circularIndex(2, 0, selectionOfIndex([2 /* TOP */, 6 /* BOTTOM */]));
    var innerSelectionSix = circularIndex(3, ANGLE_TWELFTH, selectionOfIndex([1 /* TOPLEFT */, 3 /* TOPRIGHT */, 6 /* BOTTOM */]));
    function floatColor(c) {
        return {
            r: c.r / 255,
            g: c.g / 255,
            b: c.b / 255,
            a: c.a / 255
        };
    }
    var Handlers = /** @class */ (function () {
        function Handlers(Sys, splitMode, menuSize, origin, canvas, preview) {
            this.Sys = Sys;
            this.currentMouseRelative = Point.ZERO;
            this.origin = origin;
            this.preview = preview;
            this.canvas = canvas;
            var HALF = Point.scaleConstant(0.5, menuSize);
            var BORDER_WIDTH = floor(menuSize.x * 0.03);
            var OUTER_RADIUS = floor(menuSize.x / 2) - BORDER_WIDTH;
            var MID_RADIUS = this.MID_RADIUS = floor(OUTER_RADIUS * 0.3);
            var INNER_RADIUS = this.INNER_RADIUS = floor(OUTER_RADIUS * 0.1);
            var GAP_WIDTH = floor(OUTER_RADIUS * 0.05);
            var HALF_GAP_WIDTH = floor(GAP_WIDTH / 2);
            var EDGE_WIDTH = floor(OUTER_RADIUS * 0.34);
            var CORNER_WIDTH = floor(OUTER_RADIUS * 0.4);
            var CORNER_DISTANCE = floor(OUTER_RADIUS * 0.78);
            var EDGE_ARC_DISTANCE = floor(OUTER_RADIUS * 1.1);
            var GLYPH_WIDTH = floor(OUTER_RADIUS * 0.09);
            var GLYPH_THICKNESS = floor(OUTER_RADIUS * 0.03);
            var DARK = floatColor({ r: 18, g: 36, b: 48, a: 200 });
            var LIGHT = floatColor({ r: 66, g: 79, b: 92, a: 237 });
            var BG = { luminance: 0.7, alpha: 0.7 };
            var ACTIVE = floatColor({ r: 0, g: 155, b: 203, a: 255 });
            this.selection = MenuSelection.None(splitMode);
            function setGrey(cr, grey) {
                cr.setSourceRGBA(grey.luminance, grey.luminance, grey.luminance, grey.alpha);
            }
            function setColor(cr, c) {
                cr.setSourceRGBA(c.r, c.g, c.b, c.a);
            }
            function activeColor(cr, selection, ring, location) {
                if (selection.eqPosition(ring, location)) {
                    setColor(cr, ACTIVE);
                }
                else {
                    setColor(cr, DARK);
                }
            }
            function activeColorInner(cr, selection, location) {
                activeColor(cr, selection, 1 /* INNER */, location);
            }
            function activeColorOuter(cr, selection, location) {
                activeColor(cr, selection, 2 /* OUTER */, location);
            }
            var self = this;
            this.draw = function draw(_canvas, cr, _width, _height) {
                var selection = self.selection;
                // reset surface
                cr.save();
                cr.setOperator(Sys.Cairo.Operator.CLEAR);
                cr.paint();
                cr.restore();
                // log("drawing! (radius = " + OUTER_RADIUS + ", selection = " + JSON.stringify(self.selection) + ")"); cr.save();
                // border (/backing fill)
                cr.arc(HALF.x, HALF.y, OUTER_RADIUS + BORDER_WIDTH, 0, TAO);
                setGrey(cr, BG);
                cr.fill();
                cr.save();
                cr.rectangle(0, 0, menuSize.x, menuSize.y);
                // draw everything (from now on) around the origin
                cr.translate(HALF.x, HALF.y);
                // horizontal clips: just keep drawing the same rect and rotating it..
                cr.rotate(ANGLE_SIXTEENTH);
                cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x, GAP_WIDTH);
                cr.rotate(ANGLE_EIGHTH);
                cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x, GAP_WIDTH);
                cr.rotate(ANGLE_EIGHTH);
                cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x, GAP_WIDTH);
                cr.rotate(ANGLE_EIGHTH);
                cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x, GAP_WIDTH);
                cr.rotate(ANGLE_SIXTEENTH);
                cr.clip();
                // reset rotation
                cr.rotate(PI);
                // outer fill
                cr.arc(0, 0, OUTER_RADIUS - ((OUTER_RADIUS - MID_RADIUS) / 2), 0, TAO);
                setColor(cr, LIGHT);
                cr.setLineWidth(OUTER_RADIUS - MID_RADIUS - (GAP_WIDTH / 2));
                cr.stroke();
                cr.arc(0, 0, OUTER_RADIUS, 0, TAO);
                cr.clip();
                // outer edge fills (top / left / right / bottom)
                cr.setLineWidth(EDGE_WIDTH);
                // right edge
                cr.arc(0, 0, OUTER_RADIUS - (EDGE_WIDTH / 2), -ANGLE_SIXTEENTH, ANGLE_SIXTEENTH);
                activeColorOuter(cr, selection, 4 /* RIGHT */);
                cr.stroke();
                // left edge
                cr.arc(0, 0, OUTER_RADIUS - (EDGE_WIDTH / 2), ANGLE_HALF - ANGLE_SIXTEENTH, ANGLE_HALF + ANGLE_SIXTEENTH);
                activeColorOuter(cr, selection, 0 /* LEFT */);
                cr.stroke();
                switch (selection.splitMode) {
                    case 0 /* FOUR */:
                        // bottom edge
                        cr.arc(0, 0, OUTER_RADIUS - (EDGE_WIDTH / 2), ANGLE_QUARTER - ANGLE_SIXTEENTH, ANGLE_QUARTER + ANGLE_SIXTEENTH);
                        activeColorOuter(cr, selection, 6 /* BOTTOM */);
                        cr.stroke();
                        // top edge
                        cr.arc(0, 0, OUTER_RADIUS - (EDGE_WIDTH / 2), -ANGLE_QUARTER - ANGLE_SIXTEENTH, -ANGLE_QUARTER + ANGLE_SIXTEENTH);
                        activeColorOuter(cr, selection, 2 /* TOP */);
                        cr.stroke();
                        break;
                    case 1 /* SIX */:
                        // bottom middle
                        cr.arc(0, EDGE_ARC_DISTANCE, CORNER_WIDTH, 0, TAO);
                        activeColorOuter(cr, selection, 6 /* BOTTOM */);
                        cr.fill();
                        // top middle
                        cr.arc(0, -EDGE_ARC_DISTANCE, CORNER_WIDTH, 0, TAO);
                        activeColorOuter(cr, selection, 2 /* TOP */);
                        cr.fill();
                        break;
                }
                // corner shades:
                cr.arc(CORNER_DISTANCE, CORNER_DISTANCE, CORNER_WIDTH, 0, TAO);
                activeColorOuter(cr, selection, 5 /* BOTTOMRIGHT */);
                cr.fill();
                cr.arc(-CORNER_DISTANCE, CORNER_DISTANCE, CORNER_WIDTH, 0, TAO);
                activeColorOuter(cr, selection, 7 /* BOTTOMLEFT */);
                cr.fill();
                cr.arc(-CORNER_DISTANCE, -CORNER_DISTANCE, CORNER_WIDTH, 0, TAO);
                activeColorOuter(cr, selection, 1 /* TOPLEFT */);
                cr.fill();
                cr.arc(CORNER_DISTANCE, -CORNER_DISTANCE, CORNER_WIDTH, 0, TAO);
                activeColorOuter(cr, selection, 3 /* TOPRIGHT */);
                cr.fill();
                // mid buttons:
                cr.setLineWidth(MID_RADIUS - INNER_RADIUS - HALF_GAP_WIDTH);
                cr.resetClip();
                cr.rectangle(-HALF.x, -HALF.y, menuSize.x, menuSize.y);
                switch (selection.splitMode) {
                    case 0 /* FOUR */:
                        cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x, GAP_WIDTH);
                        cr.clip();
                        cr.arc(0, 0, MID_RADIUS - ((MID_RADIUS - INNER_RADIUS) / 2) - HALF_GAP_WIDTH, PI, TAO);
                        activeColorInner(cr, selection, 2 /* TOP */);
                        cr.stroke();
                        cr.arc(0, 0, MID_RADIUS - ((MID_RADIUS - INNER_RADIUS) / 2) - HALF_GAP_WIDTH, 0, PI);
                        activeColorInner(cr, selection, 6 /* BOTTOM */);
                        cr.stroke();
                        var glyphOffset = MID_RADIUS - ((MID_RADIUS - INNER_RADIUS + GAP_WIDTH) / 2);
                        cr.rectangle(-(GLYPH_WIDTH / 2), -glyphOffset - (GLYPH_THICKNESS / 2), GLYPH_WIDTH, GLYPH_THICKNESS);
                        cr.rectangle(-(GLYPH_THICKNESS / 2), -glyphOffset - (GLYPH_WIDTH / 2), GLYPH_THICKNESS, GLYPH_WIDTH);
                        cr.rectangle(-(GLYPH_WIDTH / 2), glyphOffset - (GLYPH_THICKNESS / 2), GLYPH_WIDTH, GLYPH_THICKNESS);
                        setGrey(cr, BG);
                        cr.fill();
                        break;
                    case 1 /* SIX */:
                        cr.rotate(ANGLE_TWELFTH);
                        cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x / 2, GAP_WIDTH);
                        cr.rotate(ANGLE_TWELFTH * 4.0);
                        cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x / 2, GAP_WIDTH);
                        cr.rotate(ANGLE_TWELFTH * 4.0);
                        cr.rectangle(HALF.x, -HALF_GAP_WIDTH, -menuSize.x / 2, GAP_WIDTH);
                        cr.clip();
                        cr.rotate(ANGLE_TWELFTH * 3.0);
                        cr.arc(0, 0, MID_RADIUS - ((MID_RADIUS - INNER_RADIUS) / 2) - HALF_GAP_WIDTH, 1.5 * ANGLE_TWELFTH, 4.5 * ANGLE_TWELFTH);
                        activeColorInner(cr, selection, 6 /* BOTTOM */);
                        cr.stroke();
                        cr.arc(0, 0, MID_RADIUS - ((MID_RADIUS - INNER_RADIUS) / 2) - HALF_GAP_WIDTH, 5.0 * ANGLE_TWELFTH, 9.0 * ANGLE_TWELFTH);
                        activeColorInner(cr, selection, 1 /* TOPLEFT */);
                        cr.stroke();
                        cr.arc(0, 0, MID_RADIUS - ((MID_RADIUS - INNER_RADIUS) / 2) - HALF_GAP_WIDTH, 9.0 * ANGLE_TWELFTH, 1.0 * ANGLE_TWELFTH);
                        activeColorInner(cr, selection, 3 /* TOPRIGHT */);
                        cr.stroke();
                        break;
                }
                cr.restore();
                return Sys.Clutter.EVENT_STOP;
            };
        }
        Handlers.prototype.updateMenuSelection = function (splitMode) {
            var _a = this.currentMouseRelative, x = _a.x, y = _a.y;
            var radius = Math.sqrt(Math.pow(x, 2) + Math.pow(y, 2));
            var angle = Math.atan2(y, x);
            // log("radius = " + radius);
            // log("angle = " + angle);a
            var selection;
            if (radius <= this.INNER_RADIUS) {
                selection = MenuSelection.None(splitMode);
            }
            else if (radius < this.MID_RADIUS) {
                var innerSelection = (splitMode == 0 /* FOUR */) ? innerSelectionFour : innerSelectionSix;
                selection = MenuSelection.Inner(innerSelection(angle), splitMode);
            }
            else {
                selection = MenuSelection.Outer(outerSelection(angle), splitMode);
            }
            this.updateSelection(selection);
        };
        Handlers.prototype.onMouseMove = function (mode, splitMode, eventPoint) {
            this.currentMouseRelative = Point.subtract(eventPoint, this.origin);
            if (mode === 0 /* MENU */) {
                this.updateMenuSelection(splitMode);
            }
            this.preview.onMouseMove(mode, eventPoint);
            return this.Sys.Clutter.EVENT_STOP;
        };
        Handlers.prototype.trackMouse = function (prevMode) {
            return this.preview.trackMouse(prevMode, this.getMousePosition());
        };
        Handlers.prototype.updateSelection = function (newSelection) {
            if (!newSelection.eq(this.selection)) {
                p("updateSelection(" + JSON.stringify(newSelection) + ")");
                this.selection = newSelection;
                this.canvas.invalidate();
                this.preview.updateSelection(newSelection);
            }
        };
        Handlers.prototype.applyDirection = function (direction) {
            var currentLocation = this.selection.index;
            var selection = this.selection;
            switch (this.selection.ring) {
                case 0 /* NONE */:
                case 1 /* INNER */:
                    switch (direction) {
                        case 1 /* UP */:
                            this.updateSelection(selection.Outer(2 /* TOP */));
                            break;
                        case 3 /* DOWN */:
                            this.updateSelection(selection.Outer(6 /* BOTTOM */));
                            break;
                        case 0 /* LEFT */:
                            this.updateSelection(selection.Outer(0 /* LEFT */));
                            break;
                        case 2 /* RIGHT */:
                            this.updateSelection(selection.Outer(4 /* RIGHT */));
                            break;
                    }
                    break;
                case 2 /* OUTER */:
                    switch (direction) {
                        case 1 /* UP */:
                            switch (currentLocation) {
                                case 0 /* LEFT */:
                                    this.updateSelection(selection.Outer(1 /* TOPLEFT */));
                                    break;
                                case 4 /* RIGHT */:
                                    this.updateSelection(selection.Outer(3 /* TOPRIGHT */));
                                    break;
                                default:
                                    this.updateSelection(selection.Outer(2 /* TOP */));
                                    break;
                            }
                            break;
                        case 3 /* DOWN */:
                            switch (currentLocation) {
                                case 0 /* LEFT */:
                                    this.updateSelection(selection.Outer(7 /* BOTTOMLEFT */));
                                    break;
                                case 4 /* RIGHT */:
                                    this.updateSelection(selection.Outer(5 /* BOTTOMRIGHT */));
                                    break;
                                default:
                                    this.updateSelection(selection.Outer(6 /* BOTTOM */));
                                    break;
                            }
                            break;
                        case 0 /* LEFT */:
                            switch (currentLocation) {
                                case 2 /* TOP */:
                                    this.updateSelection(selection.Outer(1 /* TOPLEFT */));
                                    break;
                                case 6 /* BOTTOM */:
                                    this.updateSelection(selection.Outer(7 /* BOTTOMLEFT */));
                                    break;
                                default:
                                    this.updateSelection(selection.Outer(0 /* LEFT */));
                                    break;
                            }
                            break;
                        case 2 /* RIGHT */:
                            switch (currentLocation) {
                                case 2 /* TOP */:
                                    this.updateSelection(selection.Outer(3 /* TOPRIGHT */));
                                    break;
                                case 6 /* BOTTOM */:
                                    this.updateSelection(selection.Outer(5 /* BOTTOMRIGHT */));
                                    break;
                                default:
                                    this.updateSelection(selection.Outer(4 /* RIGHT */));
                                    break;
                            }
                            break;
                    }
                    break;
            }
        };
        Handlers.prototype.getSelection = function () {
            return this.selection;
        };
        Handlers.prototype.getMousePosition = function () {
            return Point.add(this.origin, this.currentMouseRelative);
        };
        return Handlers;
    }());
    MenuHandlers.Handlers = Handlers;
})(MenuHandlers || (MenuHandlers = {}));
/// <reference path="common.ts" />
/// <reference path="assert.ts" />
/// <reference path="logging.ts" />
/// <reference path="preview.ts" />
/// <reference path="point.ts" />
/// <reference path="rect.ts" />
/// <reference path="manipulations.ts" />
/// <reference path="menu_selection.ts" />
/// <reference path="menu_handlers.ts" />
var Menu;
(function (Menu_1) {
    function modeForKey(key) {
        switch (key) {
            case 64 /* CTRL */: return 2 /* MOVE */;
            case 133 /* ALT */: return 1 /* RESIZE */;
            case 65 /* SPACE */:
            case 23 /* TAB */: return 3 /* NOOP */;
            default: return null;
        }
    }
    function directionForKey(key) {
        switch (key) {
            case 45 /* K */:
            case 111 /* UP */: return 1 /* UP */;
            case 44 /* J */:
            case 116 /* DOWN */: return 3 /* DOWN */;
            case 43 /* H */:
            case 113 /* LEFT */: return 0 /* LEFT */;
            case 46 /* L */:
            case 114 /* RIGHT */: return 2 /* RIGHT */;
            default: return null;
        }
    }
    function selectionForKey(key, splitMode) {
        switch (splitMode) {
            case (0 /* FOUR */):
                switch (key) {
                    case 21 /* EQUAL */:
                        return MenuSelection.Inner(2 /* TOP */, splitMode);
                    case 20 /* MINUS */:
                        return MenuSelection.Inner(6 /* BOTTOM */, splitMode);
                }
                break;
            case (1 /* SIX */):
                switch (key) {
                    case 30 /* U */:
                        return MenuSelection.Inner(1 /* TOPLEFT */, splitMode);
                    case 31 /* I */:
                        return MenuSelection.Inner(6 /* BOTTOM */, splitMode);
                    case 32 /* O */:
                        return MenuSelection.Inner(3 /* TOPRIGHT */, splitMode);
                }
                break;
        }
        return null;
    }
    var Menu = /** @class */ (function () {
        function Menu(Sys, parent, origin, win) {
            var screen = Sys.workspaceArea(win);
            log("getting windowRect of win " + win);
            var windowRect = Sys.windowRect(win);
            assert(windowRect, "windowRect()");
            this.Sys = Sys;
            p("creating menu at " + JSON.stringify(origin) + " with bounds " + JSON.stringify(screen));
            var self = this;
            this.parent = parent;
            this.win = win;
            this.mouseMode = 0 /* MENU */;
            var backgroundActor = Sys.newClutterActor();
            backgroundActor.set_size(screen.x, screen.y);
            var menu = this.menu = Sys.newClutterActor();
            var menuSize = { x: 200, y: 200 };
            menu.set_size(menuSize.x, menuSize.y);
            var canvas = Sys.newClutterCanvas();
            canvas.set_size(menuSize.x, menuSize.y);
            menu.set_content(canvas);
            var position = Point.subtract(origin, Point.scaleConstant(0.5, menuSize));
            menu.set_position(position.x, position.y);
            this.splitMode = 0 /* FOUR */;
            var preview = this.preview = new Preview.LayoutPreview(Sys, screen, windowRect, win);
            var handlers = this.menuHandlers = new MenuHandlers.Handlers(Sys, self.splitMode, menuSize, origin, canvas, preview);
            canvas.connect('draw', handlers.draw);
            backgroundActor.connect('motion-event', function (_actor, event) {
                var position = Sys.translateEventCoordinates(Point.ofEvent(event), win);
                self.menuHandlers.onMouseMove(self.mouseMode, self.splitMode, position);
                return Sys.Clutter.EVENT_STOP;
            });
            Sys.Clutter.grab_pointer(backgroundActor);
            Sys.Clutter.grab_keyboard(backgroundActor);
            // var suspendedMouseMode = MouseMode.NOOP;
            backgroundActor.connect('key-press-event', function (_actor, event) {
                self.onKeyPress(event);
                return Sys.Clutter.EVENT_STOP;
            });
            backgroundActor.connect('key-release-event', function (_actor, event) {
                self.onKeyRelease(event);
                return Sys.Clutter.EVENT_STOP;
            });
            backgroundActor.connect('button-press-event', function () {
                self.complete(true);
                return Sys.Clutter.EVENT_STOP;
            });
            var coverPane = Sys.newClutterActor();
            coverPane.set_reactive(true);
            coverPane.connect('event', function () {
                return Sys.Clutter.EVENT_STOP;
            });
            this.ui = coverPane;
            backgroundActor.set_reactive(true);
            backgroundActor.add_actor(menu);
            coverPane.add_actor(this.preview.ui);
            coverPane.add_actor(backgroundActor);
            this.parent.insert_child_above(this.ui, null);
            backgroundActor.grab_key_focus();
            canvas.invalidate();
        }
        Menu.show = function (Sys, parent, origin, win) {
            if (!win) {
                p("no active window; ignoring");
                return null;
            }
            return new Menu(Sys, parent, origin, win);
        };
        Menu.prototype.onKeyPress = function (event) {
            var code = event.get_key_code();
            // p('keypress: ' + code);
            if (code == 9 /* ESC */) {
                this.complete(false);
            }
            else if (code == 36 /* RETURN */) {
                this.complete(true);
            }
            else if (code == 38 /* A */ && this.mouseMode == 0 /* MENU */) {
                if (this.splitMode == 0 /* FOUR */) {
                    this.splitMode = 1 /* SIX */;
                }
                else {
                    this.splitMode = 0 /* FOUR */;
                }
                this.menuHandlers.updateMenuSelection(this.splitMode);
            }
            else {
                var newMode = modeForKey(code);
                if (newMode !== null) {
                    if (this.mouseMode !== newMode) {
                        if (this.menuHandlers.trackMouse(this.mouseMode)) {
                            p("entering mode " + newMode);
                            this.mouseMode = newMode;
                            this.menu.hide();
                        }
                        else {
                            p("not entering " + newMode + " due to current menu selection");
                        }
                    }
                }
                else if (this.mouseMode == 0 /* MENU */) {
                    var direction = directionForKey(code);
                    var selection = selectionForKey(code, this.splitMode);
                    if (direction !== null) {
                        this.menuHandlers.applyDirection(direction);
                    }
                    else if (selection !== null) {
                        this.menuHandlers.updateSelection(selection);
                    }
                }
                else {
                    this.onManipulationKeyPress(code, event.get_state());
                }
            }
        };
        Menu.prototype.onManipulationKeyPress = function (code, modifiers) {
            // p("keycode: " + code);
            // p("Modifiers: " + modifiers);
            if (modifiers & this.Sys.Clutter.ModifierType.SHIFT_MASK) {
                switch (code) {
                    case 43 /* H */:
                    case 113 /* LEFT */: return this.preview.applyManipulator(Manipulations.resize(-1, "x" /* x */));
                    case 30 /* U */:
                    case 44 /* J */:
                    case 116 /* DOWN */: return this.preview.applyManipulator(Manipulations.resize(1, "y" /* y */));
                    case 31 /* I */:
                    case 45 /* K */:
                    case 111 /* UP */: return this.preview.applyManipulator(Manipulations.resize(-1, "y" /* y */));
                    case 46 /* L */:
                    case 114 /* RIGHT */: return this.preview.applyManipulator(Manipulations.resize(1, "x" /* x */));
                }
            }
            else {
                switch (code) {
                    case 43 /* H */:
                    case 113 /* LEFT */: return this.preview.applyManipulator(Manipulations.move(-1, "x" /* x */));
                    case 30 /* U */:
                    case 44 /* J */:
                    case 116 /* DOWN */: return this.preview.applyManipulator(Manipulations.move(1, "y" /* y */));
                    case 31 /* I */:
                    case 45 /* K */:
                    case 111 /* UP */: return this.preview.applyManipulator(Manipulations.move(-1, "y" /* y */));
                    case 46 /* L */:
                    case 114 /* RIGHT */: return this.preview.applyManipulator(Manipulations.move(1, "x" /* x */));
                    case 20 /* MINUS */: return this.preview.applyManipulator(Manipulations.resize(-1, null));
                    case 21 /* EQUAL */: return this.preview.applyManipulator(Manipulations.resize(1, null));
                }
            }
        };
        Menu.prototype.onKeyRelease = function (event) {
            var code = event.get_key_code();
            var fromState = modeForKey(code);
            if (fromState != null && this.mouseMode == fromState) {
                p("ending mode " + fromState);
                this.mouseMode = 3 /* NOOP */;
            }
        };
        Menu.prototype.onSelect = function (action, rect) {
            switch (action) {
                case 1 /* MINIMIZE */:
                    this.Sys.minimize(this.win);
                    return;
                case 2 /* RESIZE */:
                    this.Sys.moveResize(this.win, rect);
                    break;
                case 0 /* CANCEL */:
                    break;
            }
            this.Sys.activate(this.win);
        };
        Menu.prototype.destroy = function () {
            p("hiding menu");
            if (this.displayed()) {
                this.Sys.Clutter.ungrab_pointer();
                this.Sys.Clutter.ungrab_keyboard();
                this.parent.remove_child(this.ui);
                this.parent = null;
            }
        };
        Menu.prototype.complete = function (accept) {
            if (!accept) {
                this.onSelect(0 /* CANCEL */, null);
            }
            else {
                var selection = this.menuHandlers.getSelection();
                if (selection.eqTo(1 /* INNER */, 6 /* BOTTOM */, 0 /* FOUR */)) {
                    this.onSelect(1 /* MINIMIZE */, null);
                }
                else {
                    var rect = this.preview.getRect();
                    if (rect !== null) {
                        this.onSelect(2 /* RESIZE */, rect);
                    }
                }
            }
            this.destroy();
        };
        Menu.prototype.displayed = function () {
            return (this.parent !== null);
        };
        return Menu;
    }());
    Menu_1.Menu = Menu;
})(Menu || (Menu = {}));
/// <reference path="manipulations.ts" />
/// <reference path="assert.ts" />
/// <reference path="system.ts" />
var WindowActions;
(function (WindowActions) {
    function Make(Sys) {
        function windowManipulator(fn) {
            return function () {
                var win = Sys.currentWindow();
                if (win == null) {
                    p("no current window");
                    return;
                }
                var newRect = fn(Sys.windowRect(win), Sys.workspaceArea(win));
                Sys.moveResize(win, newRect);
                Sys.activateLater(win);
            };
        }
        function moveAction(direction, axis) {
            return windowManipulator(Manipulations.move(direction, axis));
        }
        function resizeAction(direction, axis) {
            return windowManipulator(Manipulations.resize(direction, axis));
        }
        function withWorkspaceDiff(diff, fn) {
            var workspaceIdx = Sys.workspaceIndex();
            var numWorkspaces = Sys.numWorkspaces();
            var newIdx = workspaceIdx + diff;
            if (newIdx < 0 || newIdx >= numWorkspaces)
                return;
            fn(newIdx);
        }
        function switchWorkspace(diff) {
            return function () {
                withWorkspaceDiff(diff, function (idx) {
                    Sys.activateWorkspace(idx);
                });
            };
        }
        function moveWindowWorkspace(diff) {
            return function () {
                var win = Sys.currentWindow();
                if (win === null)
                    return;
                withWorkspaceDiff(diff, function (idx) {
                    Sys.moveWindowToWorkspace(win, idx);
                });
            };
        }
        function toggleMaximize() {
            var win = Sys.currentWindow();
            if (win === null)
                return;
            if (Sys.getMaximized(win)) {
                Sys.unmaximize(win);
            }
            else {
                Sys.maximize(win);
            }
            Sys.activateLater(win);
        }
        function minimize() {
            var win = Sys.currentWindow();
            if (win === null)
                return;
            Sys.minimize(win);
        }
        function unminimize() {
            var minimized = Sys.minimizedWindows();
            // TODO: sort by last minimized?
            var win = minimized[0];
            if (!win)
                return;
            Sys.unminimize(win);
            Sys.activateLater(win);
        }
        function sizeSortOrder(rect, _screenMidpoint) {
            // negative puts biggest window first
            return -Rect.area(rect);
        }
        function radialSortOrder(rect, screenMidpoint) {
            var midpoint = Rect.midpoint(rect);
            var vector = Point.subtract(screenMidpoint, midpoint);
            var half_pi = Math.PI / 2;
            var tao = Math.PI * 2;
            var angle;
            if (Point.eqTo(vector, 0, 0)) {
                angle = -half_pi;
            }
            else {
                // atan2 gives angles in the range -PI (pointing due left) through to +PI, anti-clockwise
                angle = Math.atan2(vector.y, vector.x);
            }
            // shift angles to all be negative, then negate them to make clockwise
            angle = (angle + Math.PI);
            // take a slice on the left, just below horizontal and shift it into negative
            // so that it's ordered first
            if (angle > ((31 / 32) * tao)) {
                angle -= tao;
            }
            // p("sort order for window " + item + ":")
            // p("sort order window rect = " + JSON.stringify(item.desired_rect()) + ", midpoint = " + JSON.stringify(midpoint));
            // p("sort order angle = " + angle + ", vector = " + JSON.stringify(vector));
            // p("sort order ...");
            return angle;
        }
        var SortableWindow = /** @class */ (function () {
            function SortableWindow(win, midpoint, ordering) {
                this.win = win;
                this.order = ordering(Sys.windowRect(win), midpoint);
                this._stableSequence = null;
            }
            SortableWindow.prototype.stableSequence = function () {
                if (this._stableSequence === null) {
                    this._stableSequence = Sys.stableSequence(this.win);
                }
                return this._stableSequence;
            };
            return SortableWindow;
        }());
        function withWindowDiff(diff, fn) {
            return withWindowPair(radialSortOrder, function (i) { return i + diff; }, fn);
        }
        function withWindowPair(ordering, selectIndex, fn) {
            var _b = Sys.visibleWindows(), win = _b[0], visibleWindows = _b[1];
            if (visibleWindows.length < 2) {
                return false; // no pair
            }
            if (win === null) {
                return null; // no active window
            }
            var workArea = Sys.workspaceArea(win);
            var screenMidpoint = Point.scaleConstant(0.5, workArea);
            var windows = (visibleWindows
                .map(function (w) { return new SortableWindow(w, screenMidpoint, ordering); })
                .sort(function (a, b) {
                if (a.order === b.order) {
                    // ensure a stable sort by using index position for equivalent windows
                    return a.stableSequence() - b.stableSequence();
                }
                else {
                    return a.order - b.order;
                }
            }));
            var windowIdx = -1;
            for (var i = 0; i < windows.length; i++) {
                if (windows[i].win === win) {
                    windowIdx = i;
                    break;
                }
            }
            if (windowIdx === -1) {
                p("current window not found in visible windows");
                return null;
            }
            // p("windows (active="+ windowIdx +"): " + JSON.stringify(windows.map(function(w: SortableWindow) {
            // 	return [ Sys.windowTitle(w.win), w.order, w.stableSequence(), Rect.copy(Sys.windowRect(w.win))];
            // })));
            var newIdx = selectIndex(windowIdx);
            if (newIdx === null || newIdx === windowIdx) {
                return false;
            }
            newIdx = newIdx % windows.length;
            if (newIdx < 0)
                newIdx += windows.length;
            fn(win, windows[newIdx].win);
            return true;
        }
        function selectWindow(diff) {
            return function () {
                if (withWindowDiff(diff, function (_a, b) {
                    Sys.activate(b);
                }) === null) {
                    p("No active window, activating the first visible window");
                    var arbitraryWindow = Sys.visibleWindows()[1][0];
                    if (arbitraryWindow !== undefined) {
                        Sys.activate(arbitraryWindow);
                    }
                }
            };
        }
        function swapWindow(diff) {
            return function () {
                withWindowDiff(diff, swapWindows);
            };
        }
        function swapWindows(a, b) {
            var ar = Sys.windowRect(a);
            var br = Sys.windowRect(b);
            Sys.moveResize(b, ar);
            Sys.moveResize(a, br);
            Sys.activateLater(a); // necessary?
        }
        function swapLargestWindow() {
            return function () {
                var selectFirst = function (_currentIdx) { return 0; };
                withWindowPair(sizeSortOrder, selectFirst, swapWindows);
            };
        }
        function fillAvailableSpace() {
            log("fillAvailableSpace()");
            var _b = Sys.visibleWindows(), win = _b[0], visibleWindows = _b[1];
            if (win === null)
                return;
            var workArea = Sys.workspaceArea(win);
            var MIN_AREA = Math.min(workArea.x, workArea.y) * 20;
            var otherWindows = visibleWindows.filter(function (candidate) { return candidate !== win; });
            log("otherWindows.length = " + otherWindows.length);
            var newRect = Manipulations.largestFreeRect(workArea, otherWindows.map(Sys.windowRect), MIN_AREA);
            log("fillAvailableSpace(): found rect " + JSON.stringify(newRect));
            if (newRect !== null) {
                Sys.moveResize(win, newRect);
                Sys.activateLater(win);
            }
        }
        var RectMetrics = /** @class */ (function () {
            function RectMetrics(rect) {
                this.rect = rect;
                this.midpoint = Rect.midpoint(rect);
                this.size = Rect.area(rect);
            }
            RectMetrics.prototype.diff = function (other) {
                var posDiff = Point.magnitude(Point.subtract(this.midpoint, other.midpoint));
                var sizeDiff = Math.abs(this.size - other.size);
                // positional jumps are more jarring than size ones, so prioritize
                // close positions over sizes
                return (posDiff * 2) + sizeDiff;
            };
            return RectMetrics;
        }());
        var WindowMetrics = /** @class */ (function () {
            function WindowMetrics(win) {
                this.win = win;
                this.metrics = new RectMetrics(Sys.windowRect(win));
            }
            return WindowMetrics;
        }());
        var TileCandidate = /** @class */ (function () {
            function TileCandidate(wins, rect) {
                this.rect = rect;
                var rectMetrics = new RectMetrics(rect);
                this.match = null;
                for (var i = 0; i < wins.length; i++) {
                    var win = wins[i];
                    var diff = win.metrics.diff(rectMetrics);
                    if (this.match === null || this.diff > diff) {
                        this.diff = diff;
                        this.match = win;
                    }
                }
            }
            return TileCandidate;
        }());
        function distribute() {
            var _b = Sys.visibleWindows(), currentWindow = _b[0], windows = _b[1];
            if (windows.length === 0)
                return;
            var bounds = Sys.workspaceArea(windows[0]);
            if (windows.length === 1) {
                Sys.maximize(windows[0]);
            }
            else {
                // TODO: scatter windows by selecting the closest matches for position+size
                // instead of just distributing by index
                var leftCount_1 = Math.floor(windows.length / 2);
                var rightCount = windows.length - leftCount_1;
                var width_1 = bounds.x / 2;
                var leftHeight_1 = Math.floor(bounds.y / leftCount_1);
                var rightHeight_1 = Math.floor(bounds.y / rightCount);
                var tiles = windows.map(function (_w, i) {
                    if (i < leftCount_1) {
                        return {
                            pos: {
                                x: 0,
                                y: i * leftHeight_1
                            },
                            size: { x: width_1, y: leftHeight_1 }
                        };
                    }
                    else {
                        return {
                            pos: {
                                x: width_1,
                                y: (i - leftCount_1) * rightHeight_1
                            },
                            size: { x: width_1, y: rightHeight_1 }
                        };
                    }
                });
                var windowMetrics = windows.map(function (w) {
                    return new WindowMetrics(w);
                });
                var remainingTiles = tiles.slice();
                var remainingWindows_1 = windowMetrics.slice();
                while (remainingTiles.length > 0) {
                    // at each step, find the best match and apply it. Then remove that window & tile from
                    // candidates, and repeat
                    var candidates = remainingTiles.map(function (tile) {
                        return new TileCandidate(remainingWindows_1, tile);
                    });
                    var bestCandidate = candidates.sort(function (a, b) {
                        return a.diff - b.diff;
                    })[0];
                    // p("out of " + remainingWindows.length + ", the best candidate was '" +
                    // 	bestCandidate.match.win.get_title() + "' for rect " +
                    // 	JSON.stringify(bestCandidate.rect));
                    var windowIdx = remainingWindows_1.indexOf(bestCandidate.match);
                    var tileIdx = remainingTiles.indexOf(bestCandidate.rect);
                    assert(windowIdx !== -1);
                    assert(tileIdx !== -1);
                    remainingWindows_1.splice(windowIdx, 1);
                    remainingTiles.splice(tileIdx, 1);
                    Sys.moveResize(bestCandidate.match.win, bestCandidate.rect);
                }
            }
            if (currentWindow !== null)
                Sys.activateLater(currentWindow);
        }
        return {
            moveAction: moveAction, resizeAction: resizeAction, switchWorkspace: switchWorkspace, moveWindowWorkspace: moveWindowWorkspace,
            toggleMaximize: toggleMaximize, minimize: minimize, unminimize: unminimize, selectWindow: selectWindow, swapWindow: swapWindow,
            swapLargestWindow: swapLargestWindow, fillAvailableSpace: fillAvailableSpace, distribute: distribute
        };
    }
    WindowActions.Make = Make;
})(WindowActions || (WindowActions = {}));
/// <reference path="common.ts" />
/// <reference path="extension_settings.ts" />
/// <reference path="menu.ts" />
/// <reference path="logging.ts" />
/// <reference path="gnome_shell.ts" />
/// <reference path="actions.ts" />
var Main = imports.ui.main;
var Shell = imports.gi.Shell;
var Meta = imports.gi.Meta;
var Gdk = imports.gi.Gdk;
var Clutter = imports.gi.Clutter;
function failsafe(fn, desc) {
    return function () {
        try {
            if (desc)
                p("Running: " + desc);
            fn.apply(this, arguments);
        }
        catch (e) {
            p('Error: ' + e);
        }
    };
}
var Extension = /** @class */ (function () {
    function Extension() {
    }
    Extension.prototype.init_keybindings = function () {
        p("initializing keybindings");
        var self = this;
        var gsettings = new Settings.Keybindings().settings;
        var valid_keys = gsettings.list_keys();
        // Utility method that binds a callback to a named keypress-action.
        function handle(name, func) {
            if (valid_keys.indexOf(name) === -1) {
                // paranoia prevents a gnome shell crash
                throw (new Error("invalid key binding: " + name));
            }
            var flags = Meta.KeyBindingFlags.NONE;
            p("binding key " + name);
            var added = Main.wm.addKeybinding(name, gsettings, flags, Shell.ActionMode.NORMAL, failsafe(func, name));
            if (!added) {
                throw (new Error("failed to add keybinding handler for: " + name));
            }
            self.disable_actions.push(function () {
                p("unbinding key " + name);
                Main.wm.removeKeybinding(name);
            });
        }
        ;
        var windowActions = WindowActions.Make(GnomeSystem);
        p("adding keyboard handlers for slinger");
        handle('slinger-show', this.show_ui.bind(this));
        handle('slinger-next-window', windowActions.selectWindow(1));
        handle('slinger-prev-window', windowActions.selectWindow(-1));
        handle('slinger-swap-next-window', windowActions.swapWindow(1));
        handle('slinger-swap-prev-window', windowActions.swapWindow(-1));
        handle('slinger-swap-largest-window', windowActions.swapLargestWindow());
        handle('slinger-move-right', windowActions.moveAction(1, "x" /* x */));
        handle('slinger-move-left', windowActions.moveAction(-1, "x" /* x */));
        handle('slinger-move-up', windowActions.moveAction(-1, "y" /* y */));
        handle('slinger-move-down', windowActions.moveAction(1, "y" /* y */));
        handle('slinger-grow-horizontal', windowActions.resizeAction(1, "x" /* x */));
        handle('slinger-shrink-horizontal', windowActions.resizeAction(-1, "x" /* x */));
        handle('slinger-grow-vertical', windowActions.resizeAction(1, "y" /* y */));
        handle('slinger-shrink-vertical', windowActions.resizeAction(-1, "y" /* y */));
        handle('slinger-grow', windowActions.resizeAction(1, null));
        handle('slinger-shrink', windowActions.resizeAction(-1, null));
        handle('slinger-switch-workspace-up', windowActions.switchWorkspace(-1));
        handle('slinger-switch-workspace-down', windowActions.switchWorkspace(1));
        handle('slinger-move-window-workspace-up', windowActions.moveWindowWorkspace(-1));
        handle('slinger-move-window-workspace-down', windowActions.moveWindowWorkspace(1));
        handle('slinger-toggle-maximize', windowActions.toggleMaximize);
        handle('slinger-minimize-window', windowActions.minimize);
        handle('slinger-unminimize-window', windowActions.unminimize);
        handle('slinger-distribute', windowActions.distribute);
        handle('slinger-fill-available-space', windowActions.fillAvailableSpace);
    };
    Extension.prototype.show_ui = function () {
        this.hide_ui();
        var display = Gdk.Display.get_default();
        var window = GnomeSystem.currentWindow();
        var self = this;
        var pointer = display.get_device_manager().get_client_pointer();
        var mousePos = pointer.get_position();
        var workspaceOffset = GnomeSystem.workspaceOffset(window);
        this.menu = Menu.Menu.show(GnomeSystem, global.window_group, Point.subtract({ x: mousePos[1], y: mousePos[2] }, workspaceOffset), window);
        if (!this.menu) {
            return;
        }
        this.menu.ui.set_position(workspaceOffset.x, workspaceOffset.y);
        Main.pushModal(this.menu.ui);
        this.menu.ui.connect('unrealize', function () {
            p("hiding modal");
            GnomeSystem.setWindowHidden(window, false);
            Main.popModal(self.menu.ui);
            self.menu = null;
            return Clutter.EVENT_PROPAGATE;
        });
    };
    Extension.prototype.hide_ui = function () {
        if (this.menu) {
            this.menu.destroy();
        }
    };
    Extension.prototype.enable = function () {
        var self = this;
        failsafe(function () {
            self.disable_actions = [];
            self.init_keybindings();
            p("enabled");
        })();
    };
    Extension.prototype.disable = function () {
        var self = this;
        failsafe(function () {
            self.hide_ui();
            self.disable_actions.forEach(function (fn) {
                try {
                    fn();
                }
                catch (e) {
                    p('ERROR: ' + e);
                }
            });
            self.disable_actions = [];
            p("disabled");
        })();
    };
    return Extension;
}());
function init() {
    return new Extension();
}
